/*
    Generatore di rampe di accelerazione per motori passo-passo.
    Copyright (C) 2015 2017  Alessandro Accattini

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *****************************************************************************
 *
 *
 * $Header$
 *
 *
 *
 * REVISION HISTORY:
 * $Log$
 *
 */


/***************************************************************************/
/* include vari */

#include <stdio.h>
#include <unistd.h>
#include <math.h>

/***************************************************************************/
/* define vari */

#define UP_RAMPING	1			// in salita
#define DOWN_RAMPING	2			// in discesa

/***************************************************************************/
/* macro varie */

/***************************************************************************/
/* costanti */

static const char Licenza[]=
	"#ramping version 0.02, Copyright (C) 2015 2017  Alessandro Accattini\n";
static const char Help[]=
	"Usage: ramping total_steps\n"
	"genera la sequenza di impulsi di controllo per motori passo-passo.\n\n"
	"\t-a\tset the step angle (in radiants).\n"
	"\t-c\tset Cmin value (integer).\n"
	"\t-h\tdisplays this help.\n"
	"\t-A\tset the acceleration (default=%3.1lf rad/s2).\n"
	"\t-K\tset the initial correction factor (default=1.0).\n"
	"\t-1\tselect calculation method 1.\n"
	"\t-2\tselect calculation method 2.\n"
	"\t-3\tselect calculation method 3.\n"
	"\n";

/***************************************************************************/
/* strutture */

/***************************************************************************/
/* costanti */

#include "table.h"			// tabella coefficienti

/***************************************************************************/
/* variabili */

/***************************************************************************/
/* prototipi funzioni */

static unsigned short int FromTable(int Idx);
  /* Restituisce un elemento della tabella dei kn. */

static void Calc_1(int Step,int Cmin,double a,double alfa,double Fck,double K);
  /* Determina gli instanti di pilotaggio del motore passo-passo. */

static void Calc_2(int Step,int Cmin,double a,double alfa,double Fck,double K);
  /* Determina gli instanti di pilotaggio del motore passo-passo. */

static void Calc_3(int Step,int Cmin,double a,double alfa,double Fck,double K);
  /* Determina gli instanti di pilotaggio del motore passo-passo. */

/***************************************************************************/
/* funzioni */

static unsigned short int FromTable(int Idx)
{
  /* Restituisce un elemento della tabella dei kn. */



  if(Idx <= 0)
    return(Table[1]);			// mai minore di 1
  if(Idx >= 512)
    Idx=511;				// lo limita alla tabella
  return(Table[Idx]);			// termina
}



static void Calc_1(int Step,int Cmin,double a,double alfa,double Fck,double K)
{
  /* Determina gli instanti di pilotaggio del motore passo-passo. */



  double	c,t,Ka;
  int		MidPoint;		// punto centrale in step
  int		UpSteps=0;		// # di step necessari in salita
  int		OpMode=UP_RAMPING;	// azione in corso
  int		i=0;



  // determina c0
  t=0.0;				// instante iniziale
  MidPoint=Step>>1;			// meta' del percorso
  
  Ka=a/alfa/(Fck*Fck);			// Ka
  c=Fck * sqrt(2.0*alfa/a) * K;		// c0
  
  printf("#Ka=   %12.10lf\n",Ka);
  printf("#c0=   %4.2lf\n",c);
  printf("#i\tt\tc\n");
  printf("%5i\t%7.5lf\t%7.2lf\n",i++,t,c);
  
  // determina tutti gli istanti di commutazione del motore
  while(Step--)
  {
    t+=c/Fck;				// istante successivo


    if(OpMode == UP_RAMPING)		// durante la salita
    {
      c=c/(1.0+(Ka*c*c));		// c successivo
      
      UpSteps++;			// conta gli impulsi in salita
      if(c <= Cmin)			// raggiunta la velocita' massima
      {
	c=Cmin;				// ferma la salita della rampa
	OpMode=0;
      }
      if(UpSteps >= MidPoint)		// se e' arrivato a meta' strada
	OpMode=DOWN_RAMPING;		// passa direttamente in frenata
    }
    if(!OpMode && (Step <= UpSteps))	// se siamo prossimi all'arrivo
      OpMode=DOWN_RAMPING;		// passa in frenata
    if(OpMode == DOWN_RAMPING)		// durante la frenata
    {
      c=c/(1.0-(Ka*c*c));		// c successivo
    }
    
    
    printf("%5i\t%7.5lf\t%7.2lf\n",i++,t,c);
  }
  return;				// termina
}



static void Calc_2(int Step,int Cmin,double a,double alfa,double Fck,double K)
{
  /* Determina gli instanti di pilotaggio del motore passo-passo. */



  double	c,t,Kb;
  int		MidPoint;		// punto centrale in step
  int		UpSteps=0;		// # di step necessari in salita
  int		OpMode=UP_RAMPING;	// azione in corso
  int		i=0;
  int		n=0;



  // determina c0
  t=0.0;				// instante iniziale
  MidPoint=Step>>1;			// meta' del percorso
  
  Kb=Fck * sqrt(2.0*alfa/a);		// Kb
  c=Kb;					// c0
  n++;
  
  printf("#Kb=   %4.2lf\n",Kb);
  printf("#c0=   %4.2lf\n",c);
  printf("#i\tt\tc\n");
  printf("%5i\t%7.5lf\t%7.2lf\n",i++,t,c);
  
  // determina tutti gli istanti di commutazione del motore
  while(Step--)
  {
    t+=c/Fck;				// istante successivo


    if(OpMode == UP_RAMPING)		// durante la salita
    {
      c=Kb * (sqrt((double)(n+1)) - sqrt((double)n));
      n++;
      
      UpSteps++;			// conta gli impulsi in salita
      if(c <= Cmin)			// raggiunta la velocita' massima
      {
	c=Cmin;				// ferma la salita della rampa
	OpMode=0;
      }
      if(UpSteps >= MidPoint)		// se e' arrivato a meta' strada
	OpMode=DOWN_RAMPING;		// passa direttamente in frenata
    }
    if(!OpMode && (Step <= UpSteps))	// se siamo prossimi all'arrivo
      OpMode=DOWN_RAMPING;		// passa in frenata
    if(OpMode == DOWN_RAMPING)		// durante la frenata
    {
      c=Kb * (sqrt((double)(n+1)) - sqrt((double)n));
      if(n)
        n--;
    }

    
    printf("%5i\t%7.5lf\t%7.2lf\n",i++,t,c);
  }
  return;				// termina
}



static void Calc_3(int Step,int Cmin,double a,double alfa,double Fck,double K)
{
  /* Determina gli instanti di pilotaggio del motore passo-passo. */



  double		t;
  unsigned short int	c,Kb;
  int			MidPoint;	// punto centrale in step
  int			UpSteps=0;	// # di step necessari in salita
  int			OpMode=UP_RAMPING;	// azione in corso
  int			i=0;
  int			n=0;



  // determina c0
  t=0.0;				// instante iniziale
  MidPoint=Step>>1;			// meta' del percorso
  
  Kb=(unsigned short int)(Fck * sqrt(2.0*alfa/a));	// Kb
  c=Kb;					// c0
  n++;
  
  printf("#Kb=   %5hi\n",Kb);
  printf("#c0=   %5hi\n",c);
  printf("#i\tt\tc\n");
  printf("%5i\t%7.5lf\t%5hi\n",i++,t,c);
  
  // determina tutti gli istanti di commutazione del motore
  while(Step--)
  {
    t+=c/Fck;				// istante successivo


    if(OpMode == UP_RAMPING)		// durante la salita
    {
      c=((unsigned long int)FromTable(n) * (unsigned long int)Kb)>>16;
      n++;
      
      UpSteps++;			// conta gli impulsi in salita
      if(c <= Cmin)			// raggiunta la velocita' massima
      {
	c=Cmin;				// ferma la salita della rampa
	OpMode=0;
      }
      if(UpSteps >= MidPoint)		// se e' arrivato a meta' strada
	OpMode=DOWN_RAMPING;		// passa direttamente in frenata
    }
    if(!OpMode && (Step <= UpSteps))	// se siamo prossimi all'arrivo
      OpMode=DOWN_RAMPING;		// passa in frenata
    if(OpMode == DOWN_RAMPING)		// durante la frenata
    {
      c=((unsigned long int)FromTable(n) * (unsigned long int)Kb)>>16;
      if(n)
        n--;
    }

    
    printf("%5i\t%7.5lf\t%5hi\n",i++,t,c);
  }
  return;				// termina
}





/***************************************************************************/


int main(int argc, char **argv)
{
  void(*Method)(int,int,double,double,double,double)=NULL;
  double	a=52.0;			// accelerazione [rad/s2]
  double	alfa=3.14/12.0;		// passo del motore [rad]
  double	Fck=40e3;		// frequenza clock generatore impulsi
  double	K=1.0;			// fattore di correzione di c0
  int		Opzione;		// per la lettura delle opzioni
  int		TotalSteps;		// totale impulsi generati
  int		Cmin=100;		// ampiezza minima impulsi stepper



  // visualizza la licenza del programma
  puts(Licenza);
  
  // legge le opzioni presenti sulla riga di comando
  while((Opzione=getopt(argc,argv,"a:A:h123c:K:"))!=-1)
  {
    switch(Opzione)
    {
      case '1':
        Method=Calc_1;
        break;
      case '2':
        Method=Calc_2;
        break;
      case '3':
        Method=Calc_3;
        break;
      case 'a':
        sscanf(optarg,"%lf",&alfa);
        break;
      case 'A':
        sscanf(optarg,"%lf",&a);
        break;
      case 'K':
        sscanf(optarg,"%lf",&K);
        break;
      case 'c':
        sscanf(optarg,"%i",&Cmin);
        break;
      case 'h':
        printf(Help,a);			// stampa l'help
	return 0;			// termina
      default:
        puts("Unknown option, use -h to see help.");
	return 0;			// termina
    }
  }
  if(Method == NULL)			// se non abbiamo un metodo
  {
    puts("Please select one calculation method.");
    return 1;
  }
  
  // processa tutti gli atri parametri passati
//  for(;optind<argc;++optind)
//  {
//    printf("processing :%s \n",argv[optind]);
//  }
  if(optind >= argc)			// se sono finiti gli argomenti
  {
    printf(Help,a);			// stampa l'help
    return 0;				// termina
  }
  sscanf(argv[optind],"%i",&TotalSteps);
  printf("#Total steps: %i\n",TotalSteps);
  printf("#alfa= %9.7lf rad\n",alfa);
  printf("#a=    %3.1lf rad/s2\n",a);
  printf("#Fck=  %3.1lf Hz\n",Fck);
  printf("#Cmin= %i\n",Cmin);
  printf("#Fmax= %3.1lf Hz\n",Fck/(double)Cmin);
  printf("#K=    %3.1lf\n",K);
  
  Method(TotalSteps,Cmin,a,alfa,Fck,K);		// sequenza degli impulsi
  return 0;					// termina
}

/* $Source$ */

